#!/usr/bin/env python3
# ===-- commit-access-review.py  --------------------------------------------===#
#
# Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
# See https://llvm.org/LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# ===------------------------------------------------------------------------===#
#
# ===------------------------------------------------------------------------===#

import datetime
import github
import re
import requests
import time
import sys
import re


class User:
    THRESHOLD = 5

    def __init__(self, name, triage_list):
        self.name = name
        self.authored = 0
        self.merged = 0
        self.reviewed = 0
        self.triage_list = triage_list

    def add_authored(self, val=1):
        self.authored += val
        if self.meets_threshold():
            print(self.name, "meets the threshold with authored commits")
            del self.triage_list[self.name]

    def set_authored(self, val):
        self.authored = 0
        self.add_authored(val)

    def add_merged(self, val=1):
        self.merged += val
        if self.meets_threshold():
            print(self.name, "meets the threshold with merged commits")
            del self.triage_list[self.name]

    def add_reviewed(self, val=1):
        self.reviewed += val
        if self.meets_threshold():
            print(self.name, "meets the threshold with reviewed commits")
            del self.triage_list[self.name]

    def get_total(self):
        return self.authored + self.merged + self.reviewed

    def meets_threshold(self):
        return self.get_total() >= self.THRESHOLD

    def __repr__(self):
        return "{} : a: {} m: {} r: {}".format(
            self.name, self.authored, self.merged, self.reviewed
        )


def check_manual_requests(
    gh: github.Github, start_date: datetime.datetime
) -> list[str]:
    """
    Return a list of users who have been asked since ``start_date`` if they
    want to keep their commit access or if they have applied for commit
    access since ``start_date``
    """

    query = """
        query ($query: String!, $after: String) {
          search(query: $query, type: ISSUE, first: 100, after: $after) {
            nodes {
              ... on Issue {
                author {
                  login
                }
                body
              }
            }
            pageInfo {
              hasNextPage
              endCursor
            }
          }
        }
        """
    formatted_start_date = start_date.strftime("%Y-%m-%dT%H:%M:%S")
    variables = {
        "query": f"type:issue created:>{formatted_start_date} org:llvm repo:llvm-project label:infra:commit-access,infra:commit-access-request"
    }

    has_next_page = True
    users = []
    while has_next_page:
        res_header, res_data = gh._Github__requester.graphql_query(
            query=query, variables=variables
        )
        data = res_data["data"]
        for issue in data["search"]["nodes"]:
            users.extend([user[1:] for user in re.findall("@[^ ,\n]+", issue["body"])])
            if issue["author"]:
                users.append(issue["author"]["login"])
        has_next_page = data["search"]["pageInfo"]["hasNextPage"]
        if has_next_page:
            variables["after"] = data["search"]["pageInfo"]["endCursor"]
    return users


def get_num_commits(gh: github.Github, user: str, start_date: datetime.datetime) -> int:
    """
    Get number of commits that ``user`` has been made since ``start_date`.
    """
    variables = {
        "owner": "llvm",
        "user": user,
        "start_date": start_date.strftime("%Y-%m-%dT%H:%M:%S"),
    }

    user_query = """
        query ($user: String!) {
          user(login: $user) {
            id
          }
        }
    """

    res_header, res_data = gh._Github__requester.graphql_query(
        query=user_query, variables=variables
    )
    data = res_data["data"]
    variables["user_id"] = data["user"]["id"]

    query = """
        query ($owner: String!, $user_id: ID!, $start_date: GitTimestamp!){
          organization(login: $owner) {
            teams(query: "llvm-committers" first:1) {
              nodes {
                repositories {
                  nodes {
                    ref(qualifiedName: "main") {
                      target {
                        ... on Commit {
                          history(since: $start_date, author: {id: $user_id }) {
                            totalCount
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
     """
    count = 0
    res_header, res_data = gh._Github__requester.graphql_query(
        query=query, variables=variables
    )
    data = res_data["data"]
    for repo in data["organization"]["teams"]["nodes"][0]["repositories"]["nodes"]:
        count += int(repo["ref"]["target"]["history"]["totalCount"])
        if count >= User.THRESHOLD:
            break
    return count


def get_review_count(
    gh: github.Github, user: str, start_date: datetime.datetime
) -> int:
    """
    Return the number of reviews that ``user`` has done since ``start_date``.
    """
    query = """
        query ($query: String!) {
          search(query: $query, type: ISSUE, first: 5) {
            issueCount
          }
        }
        """
    formatted_start_date = start_date.strftime("%Y-%m-%dT%H:%M:%S")
    variables = {
        "owner": "llvm",
        "repo": "llvm-project",
        "user": user,
        "query": f"type:pr commenter:{user} -author:{user} merged:>{formatted_start_date} org:llvm",
    }

    res_header, res_data = gh._Github__requester.graphql_query(
        query=query, variables=variables
    )
    data = res_data["data"]
    return int(data["search"]["issueCount"])


def count_prs(gh: github.Github, triage_list: dict, start_date: datetime.datetime):
    """
    Fetch all the merged PRs for the project since ``start_date`` and update
    ``triage_list`` with the number of PRs merged for each user.
    """

    query = """
        query ($query: String!, $after: String) {
          search(query: $query, type: ISSUE, first: 100, after: $after) {
            issueCount,
            nodes {
              ... on PullRequest {
                 author {
                   login
                 }
                 mergedBy {
                   login
                 }
              }
            }
            pageInfo {
              hasNextPage
              endCursor
            }
          }
        }
    """
    date_begin = start_date
    date_end = None
    while date_begin < datetime.datetime.now():
        date_end = date_begin + datetime.timedelta(days=7)
        formatted_date_begin = date_begin.strftime("%Y-%m-%dT%H:%M:%S")
        formatted_date_end = date_end.strftime("%Y-%m-%dT%H:%M:%S")
        variables = {
            "query": f"type:pr is:merged merged:{formatted_date_begin}..{formatted_date_end} org:llvm",
        }
        has_next_page = True
        while has_next_page:
            print(variables)
            res_header, res_data = gh._Github__requester.graphql_query(
                query=query, variables=variables
            )
            data = res_data["data"]
            for pr in data["search"]["nodes"]:
                # Users can be None if the user has been deleted.
                if not pr["author"]:
                    continue
                author = pr["author"]["login"]
                if author in triage_list:
                    triage_list[author].add_authored()

                if not pr["mergedBy"]:
                    continue
                merger = pr["mergedBy"]["login"]
                if author == merger:
                    continue
                if merger not in triage_list:
                    continue
                triage_list[merger].add_merged()

            has_next_page = data["search"]["pageInfo"]["hasNextPage"]
            if has_next_page:
                variables["after"] = data["search"]["pageInfo"]["endCursor"]
        date_begin = date_end


def main():
    token = sys.argv[1]
    gh = github.Github(auth=github.Auth.Token(token))
    org = gh.get_organization("llvm")
    repo = org.get_repo("llvm-project")
    one_year_ago = datetime.datetime.now() - datetime.timedelta(days=365)
    triage_list = {}
    for collaborator in repo.get_collaborators(permission="push"):
        triage_list[collaborator.login] = User(collaborator.login, triage_list)

    print("Start:", len(triage_list), "triagers")
    # Step 0 Check if users have requested commit access in the last year.
    for user in check_manual_requests(gh, one_year_ago):
        if user in triage_list:
            print(user, "requested commit access in the last year.")
            del triage_list[user]
    print("After Request Check:", len(triage_list), "triagers")

    # Step 1 count all PRs authored or merged
    count_prs(gh, triage_list, one_year_ago)

    print("After PRs:", len(triage_list), "triagers")

    if len(triage_list) == 0:
        sys.exit(0)

    # Step 2 check for reviews
    for user in list(triage_list.keys()):
        review_count = get_review_count(gh, user, one_year_ago)
        triage_list[user].add_reviewed(review_count)

    print("After Reviews:", len(triage_list), "triagers")

    if len(triage_list) == 0:
        sys.exit(0)

    # Step 3 check for number of commits
    for user in list(triage_list.keys()):
        num_commits = get_num_commits(gh, user, one_year_ago)
        # Override the total number of commits to not double count commits and
        # authored PRs.
        triage_list[user].set_authored(num_commits)

    print("After Commits:", len(triage_list), "triagers")

    print("Complete:", len(triage_list), "triagers")

    with open("triagers.log", "w") as triagers_log:
        for user in triage_list:
            print(triage_list[user].__repr__())
            triagers_log.write(user + "\n")


if __name__ == "__main__":
    main()



# Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
# See https://llvm.org/LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
"""Caches .lit_test_times.txt files between premerge invocations.

.lit_test_times.txt files are used by lit to order tests to best take advantage
of parallelism. Having them around and up to date can result in a ~15%
improvement in test times. This script downloading cached test time files and
uploading new versions to the GCS buckets used for caching.
"""

import sys
import os
import logging
import multiprocessing.pool
import pathlib
import glob

from google.cloud import storage
from google.api_core import exceptions

GCS_PARALLELISM = 100


def _maybe_upload_timing_file(bucket, timing_file_path):
    if os.path.exists(timing_file_path):
        timing_file_blob = bucket.blob("lit_timing/" + timing_file_path)
        timing_file_blob.upload_from_filename(timing_file_path)


def upload_timing_files(storage_client, bucket_name: str):
    bucket = storage_client.bucket(bucket_name)
    with multiprocessing.pool.ThreadPool(GCS_PARALLELISM) as thread_pool:
        futures = []
        for timing_file_path in glob.glob("**/.lit_test_times.txt", recursive=True):
            futures.append(
                thread_pool.apply_async(
                    _maybe_upload_timing_file, (bucket, timing_file_path)
                )
            )
        for future in futures:
            future.get()
    print("Done uploading")


def _maybe_download_timing_file(blob):
    file_name = blob.name.removeprefix("lit_timing/")
    pathlib.Path(os.path.dirname(file_name)).mkdir(parents=True, exist_ok=True)
    blob.download_to_filename(file_name)


def download_timing_files(storage_client, bucket_name: str):
    bucket = storage_client.bucket(bucket_name)
    try:
        blobs = bucket.list_blobs(prefix="lit_timing")
    except exceptions.ClientError as client_error:
        print(
            "::warning file=cache_lit_timing_files.py::Failed to list blobs "
            "in bucket."
        )
        sys.exit(0)
    with multiprocessing.pool.ThreadPool(GCS_PARALLELISM) as thread_pool:
        futures = []
        for timing_file_blob in blobs:
            futures.append(
                thread_pool.apply_async(
                    _maybe_download_timing_file, (timing_file_blob,)
                )
            )
        for future in futures:
            future.wait()
            if not future.successful():
                print(
                    "::warning file=cache_lit_timing_files.py::Failed to "
                    "download lit timing file."
                )
                continue
    print("Done downloading")


if __name__ == "__main__":
    if len(sys.argv) != 2:
        logging.fatal("Expected usage is cache_lit_timing_files.py <upload/download>")
        sys.exit(1)
    action = sys.argv[1]
    storage_client = storage.Client()
    bucket_name = os.environ["CACHE_GCS_BUCKET"]
    if action == "download":
        download_timing_files(storage_client, bucket_name)
    elif action == "upload":
        upload_timing_files(storage_client, bucket_name)
    else:
        logging.fatal("Expected usage is cache_lit_timing_files.py <upload/download>")
        sys.exit(1)



# -*- Python -*-

import os
import platform
import re
import subprocess
import tempfile

import lit.formats
import lit.util

from lit.llvm import llvm_config
from lit.llvm.subst import ToolSubst
from lit.llvm.subst import FindTool

# Configuration file for the 'lit' test runner.

# name: The name of this test suite.
config.name = "BOLT"

# TODO: Consolidate the logic for turning on the internal shell by default for all LLVM test suites.
# See https://github.com/llvm/llvm-project/issues/106636 for more details.
#
# We prefer the lit internal shell which provides a better user experience on failures
# and is faster unless the user explicitly disables it with LIT_USE_INTERNAL_SHELL=0
# env var.
use_lit_shell = True
lit_shell_env = os.environ.get("LIT_USE_INTERNAL_SHELL")
if lit_shell_env:
    use_lit_shell = lit.util.pythonize_bool(lit_shell_env)

# testFormat: The test format to use to interpret tests.
#
# For now we require '&&' between commands, until they get globally killed and
# the test runner updated.
config.test_format = lit.formats.ShTest(execute_external=not use_lit_shell)

# suffixes: A list of file extensions to treat as test files.
config.suffixes = [
    ".c",
    ".cpp",
    ".cppm",
    ".m",
    ".mm",
    ".cu",
    ".ll",
    ".cl",
    ".s",
    ".S",
    ".modulemap",
    ".test",
    ".rs",
]

# excludes: A list of directories to exclude from the testsuite. The 'Inputs'
# subdirectories contain auxiliary inputs for various tests in their parent
# directories.
config.excludes = ["Inputs", "CMakeLists.txt", "README.txt", "LICENSE.txt"]

# test_source_root: The root path where tests are located.
config.test_source_root = os.path.dirname(__file__)

# test_exec_root: The root path where tests should be run.
config.test_exec_root = os.path.join(config.bolt_obj_root, "test")

# checking if maxIndividualTestTime is available on the platform and sets
# it to 60sec if so, declares lit-max-individual-test-time feature for
# further checking by tests.
supported, errormsg = lit_config.maxIndividualTestTimeIsSupported
if supported:
    config.available_features.add("lit-max-individual-test-time")
    lit_config.maxIndividualTestTime = 60
else:
    lit_config.warning(
        "Setting a timeout per test not supported. "
        + errormsg
        + " Some tests will be skipped."
    )

if config.bolt_enable_runtime:
    config.available_features.add("bolt-runtime")

if config.gnu_ld:
    config.available_features.add("gnu_ld")

if lit.util.which("fuser"):
    config.available_features.add("fuser")

llvm_config.use_default_substitutions()

llvm_config.config.environment["CLANG"] = config.bolt_clang
llvm_config.use_clang()

llvm_config.config.environment["LD_LLD"] = config.bolt_lld
ld_lld = llvm_config.use_llvm_tool("ld.lld", required=True, search_env="LD_LLD")
llvm_config.config.available_features.add("ld.lld")
llvm_config.add_tool_substitutions([ToolSubst(r"ld\.lld", command=ld_lld)])

config.substitutions.append(("%cflags", ""))
config.substitutions.append(("%cxxflags", ""))

link_fdata_cmd = os.path.join(config.test_source_root, "link_fdata.py")

tool_dirs = [config.llvm_tools_dir, config.test_source_root]

llvm_bolt_args = []

if config.libbolt_rt_instr:
    llvm_bolt_args.append(f"--runtime-instrumentation-lib={config.libbolt_rt_instr}")

if config.libbolt_rt_hugify:
    llvm_bolt_args.append(f"--runtime-hugify-lib={config.libbolt_rt_hugify}")

tools = [
    ToolSubst("llc", unresolved="fatal"),
    ToolSubst("llvm-dwarfdump", unresolved="fatal"),
    ToolSubst(
        "llvm-bolt",
        unresolved="fatal",
        extra_args=llvm_bolt_args,
    ),
    ToolSubst("llvm-boltdiff", unresolved="fatal"),
    ToolSubst("llvm-bolt-heatmap", unresolved="fatal"),
    ToolSubst("llvm-bolt-binary-analysis", unresolved="fatal"),
    ToolSubst("llvm-bat-dump", unresolved="fatal"),
    ToolSubst("perf2bolt", unresolved="fatal"),
    ToolSubst("yaml2obj", unresolved="fatal"),
    ToolSubst("llvm-mc", unresolved="fatal"),
    ToolSubst("llvm-nm", unresolved="fatal"),
    ToolSubst("llvm-objdump", unresolved="fatal"),
    ToolSubst("llvm-objcopy", unresolved="fatal"),
    ToolSubst("llvm-strings", unresolved="fatal"),
    ToolSubst("llvm-strip", unresolved="fatal"),
    ToolSubst("llvm-readelf", unresolved="fatal"),
    ToolSubst(
        "link_fdata",
        command=sys.executable,
        unresolved="fatal",
        extra_args=[link_fdata_cmd],
    ),
    ToolSubst("process-debug-line", unresolved="fatal"),
    ToolSubst("merge-fdata", unresolved="fatal"),
    ToolSubst("llvm-readobj", unresolved="fatal"),
    ToolSubst("llvm-dwp", unresolved="fatal"),
    ToolSubst("split-file", unresolved="fatal"),
]
llvm_config.add_tool_substitutions(tools, tool_dirs)


def calculate_arch_features(arch_string):
    features = []
    for arch in arch_string.split():
        features.append(arch.lower() + "-registered-target")
    return features


llvm_config.feature_config(
    [
        ("--assertion-mode", {"ON": "asserts"}),
        ("--cxxflags", {r"-D_GLIBCXX_DEBUG\b": "libstdcxx-safe-mode"}),
        ("--targets-built", calculate_arch_features),
    ]
)

config.targets = frozenset(config.targets_to_build.split(";"))
